from enum import Enum
import importlib.resources
from typing import Optional
import yaml

INFINITY = float('inf')


def load_settings(file_name: Optional[str] = None):
    if file_name:
        with open(file_name) as file:
            settings = yaml.load(file, Loader=yaml.FullLoader)
    else:
        settings = yaml.load(importlib.resources.read_text(f'{__package__}.resources', 'settings.yml'), Loader=yaml.FullLoader)
    return settings


def perc_2_float(perc: str):
    return float(perc.replace('%', '')) / 100


def month_to_year(month: int):
    return 2021 + floor(month / 12)


class Period(Enum):
    YEAR = 'year'
    MONTH = 'month'



from enum import Enum

from .common import perc_2_float


class MortgageType(Enum):
    LINEAR = 'linear'
    ANNUITY = 'annuity'


class Mortgage:
    def __init__(self, rate, size, duration, type: str):
        self._rate = perc_2_float(rate)
        self._original_size = int(size)
        self._size = int(size)
        self._duration = int(duration)
        self._type = MortgageType(type)
        self._annuity = self._calc_annuity()

    def _calc_interest(self):
        # TODO should this be different than just a simple division by the amount of months?
        return self._size * self._rate / 12

    def _calc_annuity(self):
        # https: // www.hypotheekrente - annuiteitenhypotheek.nl / annuiteit - berekenen.php
        n_periods = self._duration * 12
        monthly_interest = self._rate / 12
        return (monthly_interest / (1 - ((1 + monthly_interest) ** -n_periods))) * self._original_size

    def repay(self):
        if self._type == MortgageType.LINEAR:
            interest = self._calc_interest()
            payoff = self._original_size / self._duration / 12
            self._size -= min(payoff, self._size)
            total_payment = payoff + interest
        else:
            interest = self._calc_interest()
            payoff = self._annuity - interest
            self._size -= min(payoff, self._size)
            total_payment = self._annuity

        return total_payment, interest

    @property
    def size(self):
        return self._size



# Simulation to compare different (personal) financial scenarios
import argparse

from .asset import Asset
from .common import load_settings, Period
from .job import Job
from .mortgage import Mortgage
from .taxes import Taxes

parser = argparse.ArgumentParser(description='Run personal finance simulation.',
                                 usage="'python3 -m simulation.simulation'")


def simulate(job: Job, mortgage: Mortgage, stock_account: Asset, taxes: Taxes, settings: dict):
    nett_mortgage_payments = []
    hist = pd.DataFrame()
    for ii in range(settings['simulation']['length'] * 12):
        print(f"Month: {ii} / {settings['simulation']['length'] * 12 - 1}")
        # Income
        gross_monthly_salary = job.get_salary(month=ii)
        gross_yearly_total_salary = job.total_salary

        # TODO: properly get the holiday allowance (the job class should use the tax class for the calculation)

        # Expenses and taxes
        # gross_mortgage_payment, mortgage_interest_payment = mortgage.repay()
        # mortgage_interest_tax = taxes.calc_mortgage_interest_tax(ii, mortgage_interest_payment,
        #                                                          gross_yearly_total_salary)
        gross_mortgage_payment = 0
        mortgage_interest_tax = 0
        mortgage_interest_payment = 0

        income_tax = taxes.calc_total_income_tax(gross_yearly_total_salary, period=Period.MONTH)

        capital_gains_tax = taxes.calculate_capital_gains_tax(ii, [stock_account])

        # Determine costs
        fixed_expenses = 500

        costs_total = capital_gains_tax + gross_mortgage_payment + income_tax + fixed_expenses + mortgage_interest_tax

        # Total income
        investable_amount = gross_monthly_salary - costs_total

        # Invest
        stock_account.advance()
        stock_account.add(investable_amount)

        # Store data for visualization
        hist = hist.append({'gross_pay': gross_monthly_salary,
                            'costs_total': costs_total,
                            'capital_gains_tax': capital_gains_tax,
                            'nett_mortgage_payment': nett_mortgage_payments,
                            'mortgage_interest_payment': mortgage_interest_payment,
                            'gross_mortgage_payment': gross_mortgage_payment,
                            'income_tax': income_tax,
                            'fixed_expenses': fixed_expenses,
                            'equity': stock_account.value
                            }, ignore_index=True)

    datA = hist.copy().drop(columns=['equity'])
    datB = hist['equity']
    print(hist.head())
    datA.plot(drawstyle="steps", alpha=0.5)
    plt.show()
    datB.plot(drawstyle="steps")
    plt.title('Net worth')
    plt.xlabel('Months')
    plt.ylabel('Amount [€]')
    plt.xlim([0, 360])
    plt.ylim([0, max(datB)])
    plt.show()
    # visualize(hist)


def init_mortgage(settings):
    rate = settings['interest_rate']
    size = settings['size']
    duration = settings['duration']
    mortgage_type = settings['type']
    return Mortgage(rate, size, duration, mortgage_type)


def init_stock_investment_account(settings):
    initial_value = settings['stocks']['start_value']
    return_rate = settings['stocks']['returns']
    return Asset(initial_value, return_rate)


def main():
    parser.parse_args()
    settings = load_settings('simulation/settings.yml')
    mortgage = init_mortgage(settings['mortgage'])
    taxes = Taxes(settings['taxes'])
    job = Job(settings['salary'])
    stock_account = init_stock_investment_account(settings['assets'])
    simulate(job, mortgage, stock_account, taxes, settings)


if __name__ == '__main__':
    main()



class Asset:
    exempt_from_capital_gains_tax = False

    def __init__(self, initial_value, return_rate):
        self.value = initial_value
        self.return_rate = return_rate

    def add(self, amount):
        self.value += amount

    def advance(self):
        self.value *= 1 + self.return_rate / 12



import pathlib
from setuptools import setup, find_packages

VERSION = '0.0.8'
DESCRIPTION = 'Personal finance simulation package'

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# Setting up
setup(
    name="pfinsim",
    version=VERSION,
    author="Xander Gerrmann",
    author_email="<xander@xgerrmann.com>",
    url='https://github.com/xgerrmann/pfinsim',
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    packages=['pfinsim'],
    install_requires=[
        'PyYAML>=5.3.1'
    ],
    keywords=['python', 'simulation', 'finance', 'personal'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
    ],
    include_package_data=True,
    python_requires='>=3.7',
)



from typing import List

from .asset import Asset
from .common import month_to_year, Period, INFINITY


class Taxes:
    def __init__(self, tax_parameters):
        self.income_tax_brackets = tax_parameters['income_tax']['brackets']
        self.income_tax_rates = tax_parameters['income_tax']['rates']
        self.work_tax_brackets = tax_parameters['work_discount']['brackets']
        self.work_tax_rates = tax_parameters['work_discount']['rates']
        self.work_tax_base_amounts = tax_parameters['work_discount']['base_amounts']
        self.general_tax_discount_brackets = tax_parameters['regular_tax_discount']['brackets']
        self.general_tax_discount_base_amount = tax_parameters['regular_tax_discount']['base_amount']
        self.general_tax_discount_rates = tax_parameters['regular_tax_discount']['rates']

        # Mortgage
        self.mortgage_interest_deduction = tax_parameters['mortgage_interest_deduction']

        # Capital gains tax parameters
        self.capital_gains_tax_savings_rate = tax_parameters['capital_gains_tax']['savings_rate']
        self.capital_gains_tax_investment_rate = tax_parameters['capital_gains_tax']['investment_rate']
        self.capital_gains_tax_rate = tax_parameters['capital_gains_tax']['tax_rate']
        self.capital_gains_tax_savings_weights = tax_parameters['capital_gains_tax']['savings_weights']
        self.capital_gains_tax_brackets = tax_parameters['capital_gains_tax']['brackets']

        self.total_taxable_capital = 0

    def calc_total_income_tax(self, gross_salary_y, period: Period = Period.YEAR):
        _, income_tax = self.calc_income_tax(gross_salary_y)
        work_tax_discount = self.calc_work_tax_discount(gross_salary_y)
        general_tax_discount = self.calc_general_tax_discount(gross_salary_y)
        total_tax = income_tax - work_tax_discount - general_tax_discount
        if period == Period.MONTH:
            total_tax /= 12
        return total_tax

    def calc_highest_tax_bracket(self, gross):
        brackets = self.income_tax_brackets + [INFINITY]
        for ii, left_bracket in enumerate(brackets[:-1]):
            if left_bracket < gross < brackets[ii + 1]:
                return self.work_tax_rates[ii]

    def calc_income_tax(self, gross_y):
        nett = 0
        tax = 0
        brackets = self.income_tax_brackets + [INFINITY]
        for ii, left_bracket in enumerate(brackets[:-1]):
            rate = self.income_tax_rates[ii]
            if gross_y > left_bracket:
                right_bracket = brackets[ii + 1]
                bucket_size = min(right_bracket, gross_y) - left_bracket
                bucket_tax = bucket_size * rate
                nett += bucket_size - bucket_tax
                tax += bucket_tax
        return nett, tax

    def calc_work_tax_discount(self, gross_y):
        brackets = self.work_tax_brackets + [INFINITY]
        for ii, left_bracket in enumerate(brackets):
            rate = self.work_tax_rates[ii]
            base_amount = self.work_tax_base_amounts[ii]
            right_bracket = brackets[ii + 1]
            if left_bracket <= gross_y <= right_bracket:
                work_tax_discount_y = base_amount + rate * (gross_y - left_bracket)
                return work_tax_discount_y

    def calc_general_tax_discount(self, gross_y):
        brackets = self.general_tax_discount_brackets + [INFINITY]
        for ii, left_bracket in enumerate(brackets):
            right_bracket = brackets[ii + 1]
            rate = self.general_tax_discount_rates[ii]
            if left_bracket <= gross_y <= right_bracket:
                general_tax_discount_y = self.general_tax_discount_base_amount[ii] + rate * (gross_y - left_bracket)
                return general_tax_discount_y

    def calculate_capital_gains_tax(self, month: int, assets: List[Asset]):
        if month % 12 == 0:
            self.total_taxable_capital = self._determine_total_taxable_capital(assets)
        elif month % 12 == 3:
            return self._calc_capital_gains_tax(self.total_taxable_capital)
        return 0

    @staticmethod
    def _determine_total_taxable_capital(assets: List[Asset]):
        total_taxable_capital = 0
        for asset in assets:
            if asset.exempt_from_capital_gains_tax:
                continue
            total_taxable_capital += asset.value
        return total_taxable_capital

    def _calc_capital_gains_tax(self, total_capital: float):
        brackets = self.capital_gains_tax_brackets + [INFINITY]
        capital_gains_tax = 0
        for ii, left_bracket in enumerate(brackets[0:-1]):
            if total_capital > left_bracket:
                savings_weight = self.capital_gains_tax_savings_weights[ii]
                fictitious_return_rate = self.capital_gains_tax_savings_rate * savings_weight + \
                                         self.capital_gains_tax_investment_rate * (1 - savings_weight)
                right_bracket = brackets[ii + 1]
                bucket_size = min(total_capital, right_bracket) - left_bracket
                bucket_rate = fictitious_return_rate * self.capital_gains_tax_rate
                capital_gains_tax += bucket_rate * bucket_size
        return capital_gains_tax

    def calc_mortgage_interest_tax(self, month, interest_payment, gross_yearly_salary):
        year = month_to_year(month)
        if year in self.mortgage_interest_deduction:
            max_deduction_rate = self.mortgage_interest_deduction[year]
        else:
            max_deduction_rate = self.mortgage_interest_deduction[max(self.mortgage_interest_deduction.keys())]

        highest_income_tax_bracket = self.calc_highest_tax_bracket(gross_yearly_salary)

        mortgage_tax_deduction_rate = min(max_deduction_rate, highest_income_tax_bracket)

        mortgage_interest_tax = - mortgage_tax_deduction_rate * interest_payment
        return mortgage_interest_tax
